@@ -67,7 +67,7 @@ header_create = function(data)
                                      `Synthetic Return Numbers` = FALSE,
                                       WKT = FALSE,
                                      `Aggregate Model` = FALSE)
-  header[["Project ID - GUID"]] = uuid::UUIDgenerate()
+  header[["Project ID - GUID"]] = "00000000-0000-0000-0000-000000000000"
   header[["Version Major"]] = 1L
   header[["Version Minor"]] = 2L
   header[["System Identifier"]] = "rlas R package"
