@@ -2,9 +2,9 @@
 ===============================================================================
 
   FILE:  laspoint.hpp
-
+  
   CONTENTS:
-
+  
     This class describes an LAS point and offers helper functions to access,
     convert, and set the default (and any additional) point attributes.
 
@@ -22,13 +22,14 @@
 
     This software is distributed WITHOUT ANY WARRANTY and without even the
     implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-
+  
   CHANGE HISTORY:
-
+  
+    15 June 2018 -- fix in flag copy from legacy (0-5) to extended (6-10) type
     10 March 2017 -- fix in copy_to() and copy_from() new LAS 1.4 point types
     10 October 2016 -- small fixes for NIR and extended scanner channel
     19 July 2015 -- created after FOSS4GE in the train back from Lake Como
-
+  
 ===============================================================================
 */
 #ifndef LAS_POINT_HPP
@@ -92,11 +93,14 @@ public:
   U8 extended_return_number : 4;
   U8 extended_number_of_returns : 4;
 
-  // for 8 byte alignment of the GPS time
-  U8 dummy[3];
+  // LASlib internal use only
+  U8 deleted_flag;
 
-  // LASlib only
-  U32 deleted_flag;
+  // for 8 byte alignment of the GPS time
+  U8 dummy[2];
+
+  // compressed LASzip 1.4 points only
+  BOOL gps_time_change;
 
   F64 gps_time;
   U16 rgb[4];
@@ -190,8 +194,8 @@ public:
     }
     else if (extended_point_type)
     {
-      extended_classification = other.classification & 31;
-      extended_classification_flags = other.classification >> 5;
+      extended_classification = other.classification;
+      extended_classification_flags = ((other.withheld_flag) << 2) | ((other.keypoint_flag) << 1) | (other.synthetic_flag);
       extended_number_of_returns = other.number_of_returns;
       extended_return_number = other.return_number;
       extended_scan_angle = I16_QUANTIZE(((F32)other.scan_angle_rank)/0.006);
@@ -206,12 +210,12 @@ public:
     if (extended_point_type)
     {
       memcpy(buffer, &X, 14);
-      *((U8*)&(buffer[14])) = ((U8*)&X)[24];
-      *((U8*)&(buffer[15])) = (extended_classification_flags << 4) | (extended_scanner_channel << 2) | (((U8*)&X)[14] & 0x03);
-      *((U8*)&(buffer[16])) = extended_classification;
-      *((U8*)&(buffer[17])) = ((U8*)&X)[17];
-      *((I16*)&(buffer[18])) = *((I16*)&(((U8*)&X)[20]));
-      *((U16*)&(buffer[20])) = *((U16*)&(((U8*)&X)[18]));
+      buffer[14] = ((U8*)&X)[24]; // extended return number and number of returns
+      buffer[15] = (((U8*)&X)[14] & 0xC0) | (extended_scanner_channel << 4) | (extended_classification_flags & 0x08) | ((((U8*)&X)[15]) >> 5);
+      buffer[16] = ((U8*)&X)[23]; // extended classification
+      buffer[17] = ((U8*)&X)[17]; // user data
+      ((I16*)buffer)[9] = ((I16*)&X)[10]; // extended scan angle
+      ((U16*)buffer)[10] = ((U16*)&X)[9]; // point source ID
       memcpy(buffer+22, &gps_time, 8);
     }
     else
@@ -232,16 +236,17 @@ public:
     if (extended_point_type)
     {
       memcpy(&X, buffer, 14);
-      ((U8*)&X)[24] = *((U8*)&(buffer[14]));
-      extended_classification_flags = (*((U8*)&(buffer[15])) >> 4);
-      extended_scanner_channel = ((*((U8*)&(buffer[15])) >> 2) & 0x03);
-      scan_direction_flag = ((*((U8*)&(buffer[15])) >> 1) & 0x01);
-      edge_of_flight_line = (*((U8*)&(buffer[15])) & 0x01);
-      extended_classification = *((U8*)&(buffer[16]));
+      ((U8*)&X)[24] = buffer[14]; // extended return number and number of returns
+      extended_classification_flags = buffer[15] & 0x0F;
+      ((U8*)&X)[15] = (buffer[15] & 0x07) << 5; // legacy classification flags
+      extended_scanner_channel = (buffer[15] >> 4) & 0x03;
+      scan_direction_flag = (buffer[15] >> 6) & 0x01;
+      edge_of_flight_line = (buffer[15] >> 7) & 0x01;
+      ((U8*)&X)[23] = buffer[16]; // extended classification
       if (extended_classification < 32) classification = extended_classification;
-      ((U8*)&X)[17] = *((U8*)&(buffer[17]));
-      *((I16*)&(((U8*)&X)[20])) = *((I16*)&(buffer[18])); ;
-      *((U16*)&(((U8*)&X)[18])) = *((U16*)&(buffer[20]));
+      ((U8*)&X)[17] = buffer[17]; // user data
+      ((I16*)&X)[10] = ((I16*)buffer)[9]; // extended scan angle
+      ((U16*)&X)[9] = ((U16*)buffer)[10]; // point source ID
       memcpy(&gps_time, buffer+22, 8);
     }
     else
@@ -513,7 +518,7 @@ public:
     have_nir = FALSE;
     extra_bytes_number = 0;
     total_point_size = 0;
-
+    
     num_items = 0;
     if (items) delete [] items;
     items = 0;
@@ -553,7 +558,7 @@ public:
   inline I8 get_scan_angle_rank() const { return scan_angle_rank; };
   inline U8 get_user_data() const { return user_data; };
   inline U16 get_point_source_ID() const { return point_source_ID; };
-  inline U32 get_deleted_flag() const { return deleted_flag; };
+  inline U8 get_deleted_flag() const { return deleted_flag; };
   inline F64 get_gps_time() const { return gps_time; };
   inline const U16* get_rgb() const { return rgb; };
   inline U16 get_R() const { return rgb[0]; };
@@ -561,7 +566,7 @@ public:
   inline U16 get_B() const { return rgb[2]; };
   inline U16 get_I() const { return rgb[3]; };
   inline U16 get_NIR() const { return rgb[3]; };
-
+  
   inline void set_X(const I32 X) { this->X = X; };
   inline void set_Y(const I32 Y) { this->Y = Y; };
   inline void set_Z(const I32 Z) { this->Z = Z; };
@@ -570,14 +575,14 @@ public:
   inline void set_number_of_returns(const U8 number_of_returns) { this->number_of_returns = (number_of_returns > 7 ? 7 : number_of_returns); };
   inline void set_scan_direction_flag(const U8 scan_direction_flag) { this->scan_direction_flag = scan_direction_flag; };
   inline void set_edge_of_flight_line(const U8 edge_of_flight_line) { this->edge_of_flight_line = edge_of_flight_line; };
-  inline void set_classification(U8 classification) { this->classification = (classification & 31); };
-  inline void set_synthetic_flag(U8 synthetic_flag) { this->synthetic_flag = synthetic_flag; };
-  inline void set_keypoint_flag(U8 keypoint_flag) { this->keypoint_flag = keypoint_flag; };
-  inline void set_withheld_flag(U8 withheld_flag) { this->withheld_flag = withheld_flag; };
+  inline void set_classification(U8 classification) { if (classification < 32) { this->classification = classification; this->extended_classification = classification; } };
+  inline void set_synthetic_flag(U8 synthetic_flag) { if (synthetic_flag) { this->synthetic_flag = 1; this->extended_classification_flags |= 0x01; } else { this->synthetic_flag = 0; this->extended_classification_flags &= 0x0E; } };
+  inline void set_keypoint_flag(U8 keypoint_flag) { if (keypoint_flag) { this->keypoint_flag = 1; this->extended_classification_flags |= 0x02; } else { this->keypoint_flag = 0; this->extended_classification_flags &= 0x0D; } };
+  inline void set_withheld_flag(U8 withheld_flag) { if (withheld_flag) { this->withheld_flag = 1; this->extended_classification_flags |= 0x04; } else { this->withheld_flag = 0; this->extended_classification_flags &= 0x0B; } };
   inline void set_scan_angle_rank(I8 scan_angle_rank) { this->scan_angle_rank = scan_angle_rank; };
   inline void set_user_data(U8 user_data) { this->user_data = user_data; };
   inline void set_point_source_ID(U16 point_source_ID) { this->point_source_ID = point_source_ID; };
-  inline void set_deleted_flag(U8 deleted_flag) { this->deleted_flag = (U32)deleted_flag; };
+  inline void set_deleted_flag(U8 deleted_flag) { this->deleted_flag = deleted_flag; };
   inline void set_gps_time(const F64 gps_time) { this->gps_time = gps_time; };
   inline void set_RGB(const U16* rgb) { memcpy(this->rgb, rgb, sizeof(U16) * 3); };
   inline void set_RGBI(const U16* rgb) { memcpy(this->rgb, rgb, sizeof(U16) * 4); };
@@ -595,6 +600,8 @@ public:
   inline void set_y(const F64 y) { this->Y = quantizer->get_Y(y); };
   inline void set_z(const F64 z) { this->Z = quantizer->get_Z(z); };
 
+  inline BOOL is_extended_point_type() const { return extended_point_type; };
+
   inline U8 get_extended_classification() const { return extended_classification; };
   inline U8 get_extended_return_number() const { return extended_return_number; };
   inline U8 get_extended_number_of_returns() const { return extended_number_of_returns; };
@@ -602,7 +609,7 @@ public:
   inline U8 get_extended_overlap_flag() const { return (extended_classification_flags >> 3); };
   inline U8 get_extended_scanner_channel() const { return extended_scanner_channel; };
 
-  inline void set_extended_classification(U8 extended_classification) { this->extended_classification = extended_classification; };
+  inline void set_extended_classification(U8 extended_classification) { this->extended_classification = extended_classification; if (extended_classification > 31) this->classification = 0; else this->classification = extended_classification; };
   inline void set_extended_return_number(U8 extended_return_number) { this->extended_return_number = extended_return_number; };
   inline void set_extended_number_of_returns(U8 extended_number_of_returns) { this->extended_number_of_returns = extended_number_of_returns; };
   inline void set_extended_scan_angle(I16 extended_scan_angle) { this->extended_scan_angle = extended_scan_angle; };
@@ -612,6 +619,8 @@ public:
   inline F32 get_scan_angle() const { if (extended_point_type) return 0.006f*extended_scan_angle; else return (F32)scan_angle_rank; };
   inline F32 get_abs_scan_angle() const { if (extended_point_type) return (extended_scan_angle < 0 ? -0.006f*extended_scan_angle : 0.006f*extended_scan_angle) ; else return (scan_angle_rank < 0 ? (F32)-scan_angle_rank : (F32)scan_angle_rank); };
 
+  inline void set_scan_angle(F32 scan_angle) { if (extended_point_type) set_extended_scan_angle(I16_QUANTIZE(scan_angle/0.006f)); else set_scan_angle_rank(I8_QUANTIZE(scan_angle)); };
+
   inline void compute_coordinates()
   {
     coordinates[0] = get_x();
@@ -635,11 +644,11 @@ public:
 
   // generic functions for attributes in extra bytes
 
-  inline BOOL has_attribute(I32 index) const
+  inline BOOL has_attribute(U32 index) const
   {
     if (attributer)
     {
-      if (index < attributer->number_attributes)
+      if (((I32)index) < attributer->number_attributes)
       {
         return TRUE;
       }
@@ -647,7 +656,7 @@ public:
     return FALSE;
   };
 
-  inline BOOL get_attribute(I32 index, U8* data) const
+  inline BOOL get_attribute(U32 index, U8* data) const
   {
     if (has_attribute(index))
     {
@@ -657,7 +666,7 @@ public:
     return FALSE;
   };
 
-  inline BOOL set_attribute(I32 index, const U8* data)
+  inline BOOL set_attribute(U32 index, const U8* data) 
   {
     if (has_attribute(index))
     {
@@ -667,7 +676,7 @@ public:
     return FALSE;
   };
 
-  inline const CHAR* get_attribute_name(I32 index) const
+  inline const CHAR* get_attribute_name(U32 index) const
   {
     if (has_attribute(index))
     {
@@ -676,7 +685,7 @@ public:
     return 0;
   };
 
-  inline F64 get_attribute_as_float(I32 index) const
+  inline F64 get_attribute_as_float(U32 index) const
   {
     if (has_attribute(index))
     {
@@ -685,6 +694,14 @@ public:
     return 0.0;
   };
 
+  inline void set_attribute_as_float(U32 index, F64 value) const
+  {
+    if (has_attribute(index))
+    {
+      attributer->attributes[index].set_value_as_float(extra_bytes + attributer->attribute_starts[index], value);
+    }
+  };
+
   // typed and offset functions for attributes in extra bytes (more efficient)
 
   inline void get_attribute(I32 start, U8 &data) const { data = extra_bytes[start]; };
